----------------------------------------------------------------------
--                                                                  --
--                      ASIS_Expression_Resolver                    --
--                                                                  --
--                  Copyright (c) 2001, Marc A. Criley              --
--                                                                  --
-- ASIS_Expression_Resolver is free  software; you can redistribute --
-- it  and/or modify  it  under  terms of  the  GNU General  Public --
-- License  as published  by the  Free Software  Foundation; either --
-- version   2,   or   (at   your  option)   any   later   version. --
-- ASIS_Expression_Resolver is distributed in the hope that it will --
-- be useful,  but WITHOUT ANY  WARRANTY; without even  the implied --
-- warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. --
-- See the GNU General Public License for more details.             --
--                                                                  --
-- As a special exception, if other files instantiate generics from --
-- this unit, or you link this  unit with other files to produce an --
-- executable,  this unit does  not by  itself cause  the resulting --
-- executable  to be  covered by  the GNU  General  Public License. --
-- This exception does not however invalidate any other reasons why --
-- the executable file might be covered by the GNU Public License.  --
--                                                                  --
-- If  you have comments,  questions, have  found bugs  (and better --
-- yet,  bug fixes  :-), or  wish to  suggest  additional features, --
-- please contact the author at mcriley@acm.org.                    --
--                                                                  --
-- Change History:                                                  --
--    1.0           7 Jan 2001   Initial Release                    --
--    2.0          27 Jan 2001   Major modification to support      --
--                               float and generic discrete types   --
--                                                                  --
----------------------------------------------------------------------

with Ada.Characters.Handling;
use  Ada.Characters.Handling;
with Asis.Compilation_Units;
use  Asis.Compilation_Units;
with Asis.Declarations;
use  Asis.Declarations;
with Asis.Definitions;
use  Asis.Definitions;
with Asis.Elements;
use  Asis.Elements;
with Asis.Exceptions;
use  Asis.Exceptions;
with Asis.Expressions;
use  Asis.Expressions;

package body Asis_Expression_Resolver is

   use Asis;

   -- Exception raised when it is determined that an expression cannot
   -- be evaluated.
   Cannot_Evaluate : exception;

   -----------------------------------------------------------------------

   type Result_Reps is (Integer_Result, Float_Result);

   type Result(Result_Kind : Result_Reps := Integer_Result) is
      record
         case Result_Kind is
            when Integer_Result =>
               I_Result : Asis_Integer;
            when Float_Result =>
               F_Result : Float;
         end case;
      end record;

   -----------------------------------------------------------------------

   procedure Booleanize (L, R    : in     Integer;
                         L_Value :    out Boolean;
                         R_Value :    out Boolean) is
      -- Convert the given integers to boolean values according to
      -- whether the values are odd (1) or even (0)
   begin
      L_Value := (L mod 2) = 1;
      R_Value := (R mod 2) = 1;
   end Booleanize;

   -----------------------------------------------------------------------
   -- Overridden operators working on the operator result type.

   function "+" (L, R : Result) return Result is
   begin
      pragma Assert(L.Result_Kind = R.Result_Kind);
      case L.Result_Kind is
         when Integer_Result =>
            return (Integer_Result, L.I_Result + R.I_Result);
         when Float_Result =>
            return (Float_Result, L.F_Result + R.F_Result);
      end case;
   end "+";

   function "-" (L, R : Result) return Result is
   begin
      pragma Assert(L.Result_Kind = R.Result_Kind);
      case L.Result_Kind is
         when Integer_Result =>
            return (Integer_Result, L.I_Result - R.I_Result);
         when Float_Result =>
            return (Float_Result, L.F_Result - R.F_Result);
      end case;
   end "-";

   function "*" (L, R : Result) return Result is
   begin
      pragma Assert(L.Result_Kind = R.Result_Kind);
      case L.Result_Kind is
         when Integer_Result =>
            return (Integer_Result, L.I_Result * R.I_Result);
         when Float_Result =>
            return (Float_Result, L.F_Result * R.F_Result);
      end case;
   end "*";

   function "/" (L, R : Result) return Result is
   begin
      pragma Assert(L.Result_Kind = R.Result_Kind);
      case L.Result_Kind is
         when Integer_Result =>
            return (Integer_Result, L.I_Result / R.I_Result);
         when Float_Result =>
            return (Float_Result, L.F_Result / R.F_Result);
      end case;
   end "/";

   function "mod" (L, R : Result) return Result is
   begin
      pragma Assert(L.Result_Kind = R.Result_Kind);
      case L.Result_Kind is
         when Integer_Result =>
            return (Integer_Result, L.I_Result mod R.I_Result);
         when Float_Result =>
            raise Cannot_Evaluate;
      end case;
   end "mod";

   function "rem" (L, R : Result) return Result is
   begin
      pragma Assert(L.Result_Kind = R.Result_Kind);
      case L.Result_Kind is
         when Integer_Result =>
            return (Integer_Result, L.I_Result rem R.I_Result);
         when Float_Result =>
            raise Cannot_Evaluate;
      end case;
   end "rem";

   function "**" (L, R : Result) return Result is
   begin
      case L.Result_Kind is
         when Integer_Result =>
            -- For integral bases, the exponent cannot be negative
            if R.I_Result >= 0 then
               return (Integer_Result, L.I_Result ** R.I_Result);
            else
               raise Cannot_Evaluate;
            end if;
         when Float_Result =>
            return (Float_Result, L.F_Result ** R.I_Result);
      end case;
   end "**";

   function "+" (L : Result) return Result is
   begin
      return L;
   end "+";

   function "-" (L : Result) return Result is
   begin
      case L.Result_Kind is
         when Integer_Result =>
            return (Integer_Result, -L.I_Result);
         when Float_Result =>
            return (Float_Result, -L.F_Result);
      end case;
   end "-";

   function "abs" (L : Result) return Result is
   begin
      case L.Result_Kind is
         when Integer_Result =>
            return (Integer_Result, abs L.I_Result);
         when Float_Result =>
            return (Float_Result, abs L.F_Result);
      end case;
   end "abs";

   -- The following relational operators implicitly return a Boolean
   -- result of the operation
   function "=" (L, R : Result) return Result is
   begin
      pragma Assert(L.Result_Kind = R.Result_Kind);
      case L.Result_Kind is
         when Integer_Result =>
            return (Integer_Result, Boolean'Pos(L.I_Result = R.I_Result));
         when Float_Result =>
            return (Integer_Result, Boolean'Pos(L.F_Result = R.F_Result));
      end case;
   end "=";

   function ">" (L, R : Result) return Result is
   begin
      pragma Assert(L.Result_Kind = R.Result_Kind);
      case L.Result_Kind is
         when Integer_Result =>
            return (Integer_Result, Boolean'Pos(L.I_Result > R.I_Result));
         when Float_Result =>
            return (Integer_Result, Boolean'Pos(L.F_Result > R.F_Result));
      end case;
   end ">";

   function ">=" (L, R : Result) return Result is
   begin
      pragma Assert(L.Result_Kind = R.Result_Kind);
      case L.Result_Kind is
         when Integer_Result =>
            return (Integer_Result, Boolean'Pos(L.I_Result >= R.I_Result));
         when Float_Result =>
            return (Integer_Result, Boolean'Pos(L.F_Result >= R.F_Result));
      end case;
   end ">=";

   function "<" (L, R : Result) return Result is
   begin
      pragma Assert(L.Result_Kind = R.Result_Kind);
      case L.Result_Kind is
         when Integer_Result =>
            return (Integer_Result, Boolean'Pos(L.I_Result < R.I_Result));
         when Float_Result =>
            return (Integer_Result, Boolean'Pos(L.F_Result < R.F_Result));
      end case;
   end "<";

   function "<=" (L, R : Result) return Result is
   begin
      pragma Assert(L.Result_Kind = R.Result_Kind);
      case L.Result_Kind is
         when Integer_Result =>
            return (Integer_Result, Boolean'Pos(L.I_Result <= R.I_Result));
         when Float_Result =>
            return (Integer_Result, Boolean'Pos(L.F_Result <= R.F_Result));
      end case;
   end "<=";

   -- The following logical operations are valid only when applied to
   -- boolean expressions.  When applied to numeric expressions, the
   -- result indicates whether it came up odd (1) or even (0).
   function "not" (L : Result) return Result is
      L_Value : Boolean;
   begin
      case L.Result_Kind is
         when Integer_Result =>
            L_Value := (L.I_Result mod 2) = 1;
            return (Integer_Result, Boolean'Pos(not L_Value));
         when others =>
            raise Cannot_Evaluate;
      end case;
   end "not";

   function "and" (L, R : Result) return Result is
      L_Value : Boolean;
      R_Value : Boolean;
   begin
      case L.Result_Kind is
         when Integer_Result =>
            Booleanize(L.I_Result, R.I_Result, L_Value, R_Value);
            return (Integer_Result, Boolean'Pos(L_Value and R_Value));
         when others =>
            raise Cannot_Evaluate;
      end case;
   end "and";

   function "or" (L, R : Result) return Result is
      L_Value : Boolean;
      R_Value : Boolean;
   begin
      case L.Result_Kind is
         when Integer_Result =>
            Booleanize(L.I_Result, R.I_Result, L_Value, R_Value);
            return (Integer_Result, Boolean'Pos(L_Value or R_Value));
         when others =>
            raise Cannot_Evaluate;
      end case;
   end "or";

   function "xor" (L, R : Result) return Result is
      L_Value : Boolean;
      R_Value : Boolean;
   begin
      case L.Result_Kind is
         when Integer_Result =>
            Booleanize(L.I_Result, R.I_Result, L_Value, R_Value);
            return (Integer_Result, Boolean'Pos(L_Value xor R_Value));
         when others =>
            raise Cannot_Evaluate;
      end case;
   end "xor";


   -----------------------------------------------------------------------
   -- Note which attributes can be evaluated.  This list may increase
   -- with later releases.

   type Prefixed_Attribute_Support is array (Attribute_Kinds) of Boolean;

   Attr_Implemented : constant Prefixed_Attribute_Support :=
     (A_Pos_Attribute  |
      A_Val_Attribute  |
      A_Pred_Attribute |
      A_Succ_Attribute => True,
      others => False);

   -----------------------------------------------------------------------
   -- Forward declarations (expression evaluation is inherently
   -- recursive).
   procedure Bounds_Of_Attribute_Type (Exp   : in     Expression;
                                       Kind  : in     Result_Reps;
                                       Lower :    out Result;
                                       Upper :    out Result;
                                       Valid :    out Validity);

   function E (Exp  : Expression;
               Kind : Result_Reps) return Result;

   function Evaluate_Attribute (Exp  : Expression;
                                Kind : Result_Reps) return Result;

   procedure Evaluate_Discrete_Value_Of
     (Exp           : in     Asis.Expression;
      Discrete_Size : in     Natural;
      Value         :    out Result;
      Valid         :    out Validity);

   procedure Evaluate_Integer_Value_Of
     (Exp   : in     Asis.Expression;
      Value :    out Result;
      Valid :    out Validity);

   procedure Evaluate_Float_Value_Of
     (Exp   : in     Asis.Expression;
      Value :    out Result;
      Valid :    out Validity);

   procedure Subtype_Bounds_Of
     (Type_Decl : in     Declaration;
      Kind      : in     Result_Reps;
      Lower     :    out Result;
      Upper     :    out Result;
      Valid     :    out Validity);

   ----------------------------------------------------------------------

   function Decl_Name (Decl       : Declaration;
                       Capitalize : Boolean     := True) return String is
      -- Given an ASIS declaration, extract its name and capitalize it
      -- as requested.
      The_Name : String := To_String(Defining_Name_Image
                                     (Declarations.Names(Decl)(1)));
   begin
      if Capitalize then
         The_Name := To_Upper(The_Name);
      end if;
      return The_Name;
   end Decl_Name;

   -----------------------------------------------------------------------

   function Is_Standard_Type(Decl : Asis.Declaration) return Boolean is

      -- Return whether the declaration corresponds to one originating
      -- in package Standard.

      use type Asis.Unit_Origins;

      -- Get the declaration's compilation unit.
      Comp_Unit : Compilation_Unit := Enclosing_Compilation_Unit(Decl);

      -- Is it a predefined compilation unit?
      Predefined_Origin : Boolean := Unit_Origin(Comp_Unit) = A_Predefined_Unit;

      -- Get the declaration of that compilation unit
      Unit_Decl : Declaration := Unit_Declaration(Comp_Unit);

   begin
      -- Is the compilation unit predefined, and is it package
      -- Standard?
      return Predefined_Origin and then Decl_Name(Unit_Decl) = "STANDARD";
   end Is_Standard_Type;

   --------------------------------------------------------------------

   function Evaluate_Prefixed_Attribute(Exp : Expression;
                                        Kind : Result_Reps) return Result is
      -- Prefixed attributes are those that take an argument, such as
      -- 'First or 'Last, which operate directly on the type to which
      -- they're affixed.
      --
      -- If the prefixed attribute is an implemented one, attempt to
      -- evaluate its argument and perform the appropriate attribute
      -- operation.

      Func_Exp   : Expression := Prefix(Exp);
      Params     : Association_List := Function_Call_Parameters(Exp);
      P1         : Expression;
      Asis_Value : Result(Kind);
      Valid      : Validity;

      Attr_Kind  : Attribute_Kinds := Attribute_Kind(Func_Exp);
      -- (Not_An_Attribute returned for non-attribute function calls)

   begin
      if Attr_Implemented(Attr_Kind) then
         P1 := Actual_Parameter(Params(1));

         -- Evaluate the argument; for all the prefixed attributes
         -- implemented, the argument must be of a discrete nature.
         case Kind is
            when Integer_Result =>
               Evaluate_Integer_Value_Of(P1, Asis_Value, Valid);
            when Float_Result =>
               raise Cannot_Evaluate;
         end case;

         -- Modify the argument's value in accordance with the type of
         -- attribute
         case Attr_Kind is
            when A_Pos_Attribute |
              A_Val_Attribute =>
               -- With either of these attributes the cardinal value
               -- of the argument is returned (which was calculated
               -- above).
               null;

            -- Inappropriate usage of the Pred and Succ attributes in
            -- the source code being analyzed could result in a
            -- Constraint_Error somewhere.
            when A_Pred_Attribute =>
               case Kind is
                  when Integer_Result =>
                     Asis_Value.I_Result := Asis_Value.I_Result - 1;
                  when Float_Result =>
                     raise Cannot_Evaluate;
               end case;

            when A_Succ_Attribute =>
               case Kind is
                  when Integer_Result =>
                     Asis_Value.I_Result := Asis_Value.I_Result + 1;
                  when Float_Result =>
                     raise Cannot_Evaluate;
               end case;
            when others =>
               raise Cannot_Evaluate;
         end case;
         return Asis_Value;
      else
         raise Cannot_Evaluate;
      end if;
   end Evaluate_Prefixed_Attribute;

   --------------------------------------------------------------------

   function Evaluate_Function_Call (Exp  : Expression;
                                    Kind : Result_Reps) return Result is
      -- Evaluate a function call. For AER purposes, the only
      -- acceptable function calls are those associated with the
      -- predefined attributes and operators.
      Function_Expression : Expression;
      Operator : Operator_Kinds;
      Params   : Association_List := Function_Call_Parameters(Exp);
      P1, P2 : Expression;

   begin

      if Is_Prefix_Call(Exp) then
         -- Attributes with arguments (e.g., 'Pred, 'Succ)
         return Evaluate_Prefixed_Attribute(Exp, Kind);
      else
         -- Infix (predefined) operator
         Function_Expression := Prefix(Exp);
         pragma Assert(Expression_Kind(Function_Expression) = An_Operator_Symbol);

         -- Extract the parameters
         pragma Assert(Association_Kind(Params(1)) = A_Parameter_Association);

         -- Extract the first parameter
         P1 := Actual_Parameter(Params(1));

         -- If there's a second parameter, get it as well.  In the
         -- operator evaluation itself it will be asserted that the
         -- correct number of parameters has been supplied.  (Note
         -- that only syntactically and semantically correct code can
         -- be properly processed with ASIS, so there should never be
         -- an inconsistency in this area.)
         if Params'Length > 1 then
            P2 := Actual_Parameter(Params(2));
         end if;

         -- Evaluate the expression
         Operator := Operator_Kind(Function_Expression);

         case Operator is
            when A_Plus_Operator =>
               pragma Assert(Params'Length = 2);
               return E(P1, Kind) + E(P2, Kind);
            when A_Minus_Operator =>
               pragma Assert(Params'Length = 2);
               return E(P1, Kind) - E(P2, Kind);
            when A_Multiply_Operator =>
               pragma Assert(Params'Length = 2);
               return E(P1, Kind) * E(P2, Kind);
            when A_Divide_Operator =>
               pragma Assert(Params'Length = 2);
               return E(P1, Kind) / E(P2, Kind);
            when A_Mod_Operator =>
               pragma Assert(Params'Length = 2);
               return E(P1, Kind) mod E(P2, Kind);
            when A_Rem_Operator =>
               pragma Assert(Params'Length = 2);
               return E(P1, Kind) rem E(P2, Kind);
            when An_Exponentiate_Operator =>
               pragma Assert(Params'Length = 2);
               -- For exponentiation, the right operand must be of any
               -- integer type
               return E(P1, Kind) ** E(P2, Integer_Result);
            when A_Unary_Plus_Operator =>
               pragma Assert(Params'Length = 1);
               return E(P1, Kind);
            when A_Unary_Minus_Operator =>
               pragma Assert(Params'Length = 1);
               return -E(P1, Kind);
            when An_Abs_Operator =>
               pragma Assert(Params'Length = 1);
               return abs(E(P1, Kind));
            when A_Not_Operator =>
               pragma Assert(Params'Length = 1);
               case Kind is
                  when Integer_Result =>
                     return not(E(P1, Kind));
                  when others =>
                     raise Cannot_Evaluate;
               end case;
            when An_And_Operator =>
               pragma Assert(Params'Length = 2);
               case Kind is
                  when Integer_Result =>
                     return E(P1, Kind) and E(P2, Kind);
                  when others =>
                     raise Cannot_Evaluate;
               end case;
            when An_Or_Operator =>
               pragma Assert(Params'Length = 2);
               case Kind is
                  when Integer_Result =>
                     return E(P1, Kind) or E(P2, Kind);
                  when others =>
                     raise Cannot_Evaluate;
               end case;
            when An_Xor_Operator =>
               pragma Assert(Params'Length = 2);
               case Kind is
                  when Integer_Result =>
                     return E(P1, Kind) xor E(P2, Kind);
                  when others =>
                     raise Cannot_Evaluate;
               end case;
            when An_Equal_Operator =>
               pragma Assert(Params'Length = 2);
               case Kind is
                  when Integer_Result =>
                     return E(P1, Kind) = E(P2, Kind);
                  when others =>
                     raise Cannot_Evaluate;
               end case;
            when A_Not_Equal_Operator =>
               pragma Assert(Params'Length = 2);
               case Kind is
                  when Integer_Result =>
                     return not(E(P1, Kind) = E(P2, Kind));
                  when others =>
                     raise Cannot_Evaluate;
               end case;
            when A_Less_Than_Operator =>
               pragma Assert(Params'Length = 2);
               return E(P1, Kind) < E(P2, Kind);
            when A_Less_Than_Or_Equal_Operator =>
               pragma Assert(Params'Length = 2);
               return E(P1, Kind) <= E(P2, Kind);
            when A_Greater_Than_Operator =>
               pragma Assert(Params'Length = 2);
               return E(P1, Kind) > E(P2, Kind);
            when A_Greater_Than_Or_Equal_Operator =>
               pragma Assert(Params'Length = 2);
               return E(P1, Kind) >= E(P2, Kind);
            when others =>
               raise Cannot_Evaluate;
         end case;
      end if;
   end Evaluate_Function_Call;

   ---------------------------------------------------------

   function Evaluate_Identifier (Exp  : Expression;
                                 Kind : Result_Reps) return Result is
      Id : Element;
      Value_Expression : Expression;
      Constant_Definition : Defining_Name;
   begin
      -- Retrieve the declaration associated with the identifier
      -- expression.  This is embedded within an exception trapping
      -- block because some kinds of identifiers do not have
      -- corresponding declarations, and so ASIS_Inappropriate_Element
      -- is raised when an attempt is made to retrieve that
      -- declaration.  Unfortunately ASIS provides no straightforward
      -- way to check for this situation beforehand.
      begin
         Id := Corresponding_Name_Declaration (Exp);
      exception
         when ASIS_Inappropriate_Element =>
            raise Cannot_Evaluate;
      end;

      pragma Assert(Element_Kind(Id) = A_Declaration);

      -- Process according to the type of declaration
      case Declaration_Kind(Id) is
         when A_Constant_Declaration =>
            Value_Expression := Initialization_Expression(Id);
            return E(Value_Expression, Kind);

         when A_Deferred_Constant_Declaration =>
            -- With a deferred constant, go over and get the actual
            -- constant definition and get its value.
            Id := Corresponding_Name_Definition (Exp);
            pragma Assert(Element_Kind(Id) = A_Defining_Name);
            Constant_Definition := Corresponding_Constant_Declaration(Id);
            Value_Expression := Initialization_Expression(Constant_Definition);
            return E(Value_Expression, Kind);

         when An_Integer_Number_Declaration =>
            Value_Expression := Initialization_Expression(Id);
            return E(Value_Expression, Kind);

         when A_Variable_Declaration =>
            -- Evaluate the initialization expression, if present, of
            -- a variable declaration
            Value_Expression := Initialization_Expression(Id);
            if Is_Equal
              (Value_Expression, Nil_Element) then
               raise Cannot_Evaluate;
            else
               return E(Value_Expression, Kind);
            end if;

         when others =>
            raise Cannot_Evaluate;
      end case;
   end Evaluate_Identifier;

   ---------------------------------------------------------

   procedure Evaluate_Constraint
     (Type_Constraint : in     Constraint;
      Kind            : in     Result_Reps;
      Lower           :    out Result;
      Upper           :    out Result;
      Valid           :    out Validity) is

      -- Given a range constraint, extract its high and low bounds.

      Range_Reference : Expression;
      Range_Target    : Expression;

   begin
      case Constraint_Kind(Type_Constraint) is
         when A_Range_Attribute_Reference =>
            Range_Reference := Range_Attribute(Type_Constraint);
            Range_Target := Prefix(Range_Reference);
            Bounds_Of_Attribute_Type(Range_Target, Kind, Lower, Upper, Valid);

         when A_Simple_Expression_Range =>
            Lower := E(Lower_Bound(Type_Constraint), Kind);
            Upper := E(Upper_Bound(Type_Constraint), Kind);
            Valid := Resolved;

         when others =>
            raise Cannot_Evaluate;
      end case;
   end Evaluate_Constraint;

   ---------------------------------------------------------

   procedure Bounds_Of
     (Type_Decl : in     Asis.Declaration;
      Type_Defn : in     Asis.Definition;
      Kind      : in     Result_Reps;
      Lower     :    out Result;
      Upper     :    out Result;
      Valid     :    out Validity) is

      -- Extract the high and low bounds associated with a
      -- declaration.  (This procedure, of course, may only be invoked
      -- with a declaration that has such bounds.)

      Defn_Type_Kind : Type_Kinds;
      Bounds_Constraint : Range_Constraint;
      Lower_Expr : Expression;
      Upper_Expr : Expression;

      Type_Decl_Kind : Declaration_Kinds;

      Parent_Subtype : Subtype_Indication;
      Constraint_Decl : Declaration;
      Constraint_Defn : Definition;

      Type_Constraint : Constraint;

      Parent_Subtype_Expr : Expression;
      Parent_Subtype_Decl : Declaration;
      Parent_Subtype_Defn : Definition;

   begin
      Defn_Type_Kind := Type_Kind(Type_Defn);

      -- Process according to the type of definition being processed.

      case Defn_Type_Kind is
         when A_Signed_Integer_Type_Definition =>
            Bounds_Constraint := Integer_Constraint(Type_Defn);
            Lower_Expr := Lower_Bound(Bounds_Constraint);
            Upper_Expr := Upper_Bound(Bounds_Constraint);

            case Kind is
               when Integer_Result =>
                  Evaluate_Integer_Value_Of(Lower_Expr, Lower, Valid);
                  if Valid = Resolved then
                     Evaluate_Integer_Value_Of(Upper_Expr, Upper, Valid);
                  end if;
               when Float_Result =>
                  raise Cannot_Evaluate;
            end case;

         when A_Modular_Type_Definition =>
            Upper_Expr := Mod_Static_Expression(Type_Defn);

            case Kind is
               when Integer_Result =>
                  Lower := (Result_Kind => Integer_Result, I_Result => 0);
                  Evaluate_Integer_Value_Of(Upper_Expr, Upper, Valid);
                  Upper.I_Result := Upper.I_Result - 1;
               when Float_Result =>
                  raise Cannot_Evaluate;
            end case;

         when A_Derived_Type_Definition =>
            -- Get the parent subtype
            Parent_Subtype := Parent_Subtype_Indication(Type_Defn);
            Type_Constraint := Subtype_Constraint(Parent_Subtype);

            if Is_Equal(Type_Constraint, Nil_Element) then
               -- While the derived type definition references
               -- A_Subtype_Indication, there was no constraint placed
               -- on it.  Therefore ascend via the type name.
               Parent_Subtype_Expr := Definitions.Subtype_Mark(Parent_Subtype);
               Parent_Subtype_Decl := Corresponding_Name_Declaration(Parent_Subtype_Expr);
               pragma Assert(Element_Kind(Parent_Subtype_Decl) = A_Declaration);

               Type_Decl_Kind := Declaration_Kind(Parent_Subtype_Decl);
               if Type_Decl_Kind = A_Subtype_Declaration then
                  Subtype_Bounds_Of(Parent_Subtype_Decl, Kind, Lower, Upper, Valid);
               else
                  Parent_Subtype_Defn := Type_Declaration_View(Parent_Subtype_Decl);
                  Bounds_Of(Parent_Subtype_Decl, Parent_Subtype_Defn,
                            Kind, Lower, Upper, Valid);
               end if;

            else
               Evaluate_Constraint(Type_Constraint, Kind, Lower, Upper, Valid);
            end if;

         when An_Enumeration_Type_Definition =>
            declare
               -- Retrieve the enumeration literal specifications...
               Enum_Decl_List : Declaration_List :=
                 Enumeration_Literal_Declarations(Type_Defn);
               Lower_Defining_Name : Defining_Name_List(1..1);
               Upper_Defining_Name : Defining_Name_List(1..1);
            begin
               -- ...then extract the 'Pos value of the first and last
               -- ones:

               -- Lower
               Lower_Defining_Name := Names(Enum_Decl_List(Enum_Decl_List'First));

               -- Upper
               Upper_Defining_Name := Names(Enum_Decl_List(Enum_Decl_List'Last));

               case Kind is
                  when Integer_Result =>
                     Lower := (Result_Kind => Integer_Result,
                               I_Result    =>
                                 Asis_Natural'Wide_Value(Position_Number_Image
                                                         (Lower_Defining_Name(1))));
                     Upper := (Result_Kind => Integer_Result,
                               I_Result    =>
                                 Asis_Natural'Wide_Value(Position_Number_Image
                                                         (Upper_Defining_Name(1))));
                  when Float_Result =>
                     raise Cannot_Evaluate;
               end case;

               Valid := Resolved;
            end;

         when A_Floating_Point_Definition |
           An_Ordinary_Fixed_Point_Definition |
           A_Decimal_Fixed_Point_Definition =>
            Bounds_Constraint := Real_Range_Constraint(Type_Defn);

            if Is_Equal(Bounds_Constraint, Nil_Element) then
               if Is_Standard_Type(Type_Decl) then
                  -- This is Standard.Float
                  Lower := (Float_Result, Float'First);
                  Upper := (Float_Result, Float'Last);
                  Valid := Resolved;
               else
                  raise Cannot_Evaluate;
               end if;
            else
               Lower_Expr := Lower_Bound(Bounds_Constraint);
               Upper_Expr := Upper_Bound(Bounds_Constraint);

               Evaluate_Float_Value_Of(Lower_Expr, Lower, Valid);
               if Valid = Resolved then
                  Evaluate_Float_Value_Of(Upper_Expr, Upper, Valid);
               end if;
            end if;

         when others =>
            Valid := No_Value;
      end case;
   end Bounds_Of;

   ---------------------------------------------------------

   procedure Subtype_Bounds_Of
     (Type_Decl : in     Declaration;
      Kind      : in     Result_Reps;
      Lower     :    out Result;
      Upper     :    out Result;
      Valid     :    out Validity) is

      -- Extract the high and low bounds of a subtype

      Type_Defn       : Asis.Definition := Type_Declaration_View(Type_Decl);
      Type_Constraint : Constraint;
      Constraint_Decl : Asis.Declaration;
      Constraint_Decl_Kind : Declaration_Kinds;

   begin
      pragma Assert(Definition_Kind(Type_Defn) = A_Subtype_Indication);

      -- If specified, get and process the constraint associated with
      -- this subtype.
      Type_Constraint := Subtype_Constraint(Type_Defn);

      if Is_Equal(Type_Constraint, Nil_Element) then
         -- There was no explicit constraint on the subtype, so go up
         -- and try to find one.
         Constraint_Decl := Corresponding_Last_Constraint(Type_Decl);
         Type_Defn := Type_Declaration_View(Constraint_Decl);
         Constraint_Decl_Kind := Declaration_Kind(Constraint_Decl);

         if Constraint_Decl_Kind = A_Subtype_Declaration then
            -- Since Corresponding_Last_Constraint ended up with a
            -- A_Subtype_Declaration, it _has_ to have a constraint!
            Type_Constraint := Subtype_Constraint(Type_Defn);
            Evaluate_Constraint(Type_Constraint, Kind, Lower, Upper, Valid);
         elsif Constraint_Decl_Kind = An_Ordinary_Type_Declaration then
            -- An_Ordinary_Type_Declaration can be processed by just
            -- going out and getting its bounds
            Bounds_Of(Type_Decl, Type_Defn, Kind, Lower, Upper, Valid);
         else
            raise Cannot_Evaluate;
         end if;
      else
         Evaluate_Constraint(Type_Constraint, Kind, Lower, Upper, Valid);
      end if;
   end Subtype_Bounds_Of;

   ---------------------------------------------------------

   procedure Bounds_Of_Attribute_Type (Exp   : in     Expression;
                                       Kind  : in     Result_Reps;
                                       Lower :    out Result;
                                       Upper :    out Result;
                                       Valid :    out Validity) is

      -- A type declaration has a 'Range attribute applied to it, so
      -- do the necessary processing to extract the bounds of that
      -- type's range.

      Attribute_Target : Expression := Exp;

      Type_Decl : Declaration;
      Decl_Kind : Declaration_Kinds;

      Type_Defn : Definition;
      Type_Defn_Kind : Definition_Kinds;

   begin
      if Expression_Kind(Exp) = A_Selected_Component then
         -- Get just the type element
         Attribute_Target := Selector(Attribute_Target);
      end if;

      pragma Assert(Expression_Kind(Attribute_Target) = An_Identifier);

      -- Get the type definition upon which the attribute is being
      -- applied
      Type_Decl := Corresponding_Name_Declaration(Attribute_Target);

      Decl_Kind := Declaration_Kind(Type_Decl);
      if (Decl_Kind /= An_Ordinary_Type_Declaration) and
        (Decl_Kind /= A_Subtype_Declaration) then
         raise Cannot_Evaluate;
      end if;

      Type_Defn := Type_Declaration_View(Type_Decl);

      Type_Defn_Kind := Definition_Kind(Type_Defn);

      Valid := No_Value;
      if Type_Defn_Kind = A_Type_Definition then
         -- Get the bounds for a type definition
         Bounds_Of(Type_Decl, Type_Defn, Kind, Lower, Upper, Valid);

      elsif Type_Defn_Kind = A_Subtype_Indication then
         Subtype_Bounds_Of(Type_Decl, Kind, Lower, Upper, Valid);
      else
         raise Cannot_Evaluate;
      end if;

      if Valid /= Resolved then
         raise Cannot_Evaluate;
      end if;

   end Bounds_Of_Attribute_Type;

   ---------------------------------------------------------

   function Evaluate_Attribute (Exp  : Expression;
                                Kind : Result_Reps) return Result is
      -- Handle the processing of a 'First or 'Last attribute

      Attr_Kind : Attribute_Kinds := Attribute_Kind(Exp);

      -- Initialize with the prefix of the attribute.  Whether the
      -- attribute can or cannot be evaluated, there has to be
      -- _something_ in front of the tick mark.
      Attribute_Target : Expression := Prefix(Exp);

      Lower : Result(Kind);
      Upper : Result(Kind);

      Valid : Validity;

   begin
      if (Attr_Kind = A_First_Attribute) or (Attr_Kind = A_Last_Attribute) then
         Bounds_Of_Attribute_Type(Attribute_Target, Kind, Lower, Upper, Valid);
         case Attr_Kind is
            when A_First_Attribute =>
               return Lower;
            when A_Last_Attribute =>
               return Upper;
            when others =>
               raise Cannot_Evaluate;
         end case;
      else
         raise Cannot_Evaluate;
      end if;
   end Evaluate_Attribute;

   ---------------------------------------------------------

   function Evaluate_Type_Conversion (Exp  : Expression;
                                      Kind : Result_Reps) return Result is
      -- Process a type conversion.  If the expected kind matches the
      -- actual kind, just return the resolved value, otherwise do a
      -- type conversion.
      Arg_Result : Result;
   begin
      -- Evaluate the argument, which may or may not be of the desired
      -- result kind
      Arg_Result := E(Converted_Or_Qualified_Expression(Exp), Kind);
      if Arg_Result.Result_Kind = Kind then
         -- The actual and desired types match, return the result
         return Arg_Result;
      else
         if Kind = Integer_Result then
            return (Integer_Result, Integer(Arg_Result.F_Result));
         else
            return (Float_Result, Float(Arg_Result.I_Result));
         end if;
      end if;
   end Evaluate_Type_Conversion;

   ---------------------------------------------------------

   function E (Exp  : Expression;
               Kind : Result_Reps) return Result is
      Expr_Kind : Expression_Kinds := Expression_Kind(Exp);
   begin
      case Expr_Kind is
         when An_Integer_Literal =>
            return (Result_Kind => Integer_Result,
                    I_Result    => Asis_Natural'Wide_Value(Value_Image(Exp)));
         when An_Enumeration_Literal | A_Character_Literal =>
            case Kind is
               when Integer_Result =>
                  return (Result_Kind => Integer_Result,
                          I_Result    => Asis_Natural'Wide_Value
                          (Position_Number_Image(Corresponding_Name_Definition(Exp))));
               when Float_Result =>
                  raise Cannot_Evaluate;
            end case;
         when A_Real_Literal =>
            return (Result_Kind => Float_Result,
                    F_Result    => Float'Wide_Value(Value_Image(Exp)));
         when An_Identifier =>
            return Evaluate_Identifier(Exp, Kind);
         when A_Selected_Component =>
            return E(Selector(Exp), Kind);
         when A_Parenthesized_Expression =>
            return E(Expression_Parenthesized(Exp), Kind);
         when A_Type_Conversion =>
            return Evaluate_Type_Conversion(Exp, Kind);
         when A_Qualified_Expression =>
            return E(Converted_Or_Qualified_Expression(Exp), Kind);
         when An_Attribute_Reference =>
            return Evaluate_Attribute(Exp, Kind);
         when A_Function_Call =>
            return Evaluate_Function_Call(Exp, Kind);
         when others =>
            raise Cannot_Evaluate;
      end case;
   end E;

   --------------------------------
   -- Evaluate_Discrete_Value_Of --
   --------------------------------

   procedure Evaluate_Discrete_Value_Of
     (Exp           : in     Asis.Expression;
      Discrete_Size : in     Natural;
      Value         :    out Result;
      Valid         :    out Validity) is
      -- Initiate the evaluation of an expression as a discrete-kind
      -- of entity.  Handle any exceptions raised as a result of that
      -- evaluation and convert them to status values.

      Asis_Value : Result(Integer_Result);

   begin
      -- Ensure this is an expression
      if Element_Kind(Exp) = An_Expression then
         Asis_Value := E(Exp, Integer_Result);
         Value := Asis_Value;
         Valid := Resolved;
      else
         Valid := Improper_Element;
      end if;
   exception
      when Cannot_Evaluate =>
         Valid := No_Value;
      when Asis_Failed =>
         Valid := Asis_Failure;
      when Constraint_Error =>
         -- An attempt was made to process an expression whose value
         -- was outside the bounds of Integer, e.g., a modular typed
         -- expression was a large value.
         Valid := Out_Of_Range;
   end Evaluate_Discrete_Value_Of;

   -------------------------------
   -- Evaluate_Integer_Value_Of --
   -------------------------------

   procedure Evaluate_Integer_Value_Of
     (Exp   : in     Asis.Expression;
      Value :    out Result;
      Valid :    out Validity)
   is
      -- Initiate the evaluation of an expression as an integer-kind
      -- of entity.  Handle any exceptions raised as a result of that
      -- evaluation and convert them to status values.
      Asis_Value : Result;
   begin
      -- Ensure this is an expression
      if Element_Kind(Exp) = An_Expression then
         Asis_Value := E(Exp, Integer_Result);
         case Asis_Value.Result_Kind is
            when Integer_Result =>
               Value := Asis_Value;
            when Float_Result =>
               Value.I_Result := Integer(Asis_Value.F_Result);
         end case;
         Valid := Resolved;
      else
         Valid := Improper_Element;
      end if;
   exception
      when Cannot_Evaluate =>
         Valid := No_Value;
      when Asis_Failed =>
         Valid := Asis_Failure;
      when Constraint_Error =>
         -- An attempt was made to process an expression whose value
         -- was outside the bounds of Integer, i.e., a modular typed
         -- expression was a large value.
         Valid := Out_Of_Range;
   end Evaluate_Integer_Value_Of;

   ----------------------
   -- Integer_Value_Of --
   ----------------------

   procedure Integer_Value_Of
     (Exp   : in     Asis.Expression;
      Value :    out Integer;
      Valid :    out Validity)
   is
      Asis_Value : Result(Integer_Result);
   begin
      Evaluate_Integer_Value_Of(Exp, Asis_Value, Valid);
      if Valid = Resolved then
         Value := Asis_Value.I_Result;
      end if;
   end Integer_Value_Of;

   -----------------------
   -- Cardinal_Value_Of --
   -----------------------

   procedure Cardinal_Value_Of (Exp   : in     Asis.Expression;
                                Value :    out Cardinal_Type;
                                Valid :    out Validity) is
      Cardinal_Value : Integer;
   begin
      Integer_Value_Of(Exp, Cardinal_Value, Valid);
      if Valid = Resolved then
         Value := Cardinal_Type'Val(Cardinal_Value);
      end if;
   exception
      when Cannot_Evaluate =>
         Valid := No_Value;
      when Asis_Failed =>
         Valid := Asis_Failure;
      when Constraint_Error =>
         Valid := Out_Of_Range;
   end Cardinal_Value_Of;

   -----------------------------
   -- Evaluate_Float_Value_Of --
   -----------------------------

   procedure Evaluate_Float_Value_Of
     (Exp   : in     Asis.Expression;
      Value :    out Result;
      Valid :    out Validity)
   is
      -- Initiate the evaluation of an expression as a floating
      -- point-kind of entity.  Handle any exceptions raised as a
      -- result of that evaluation and convert them to status values.
      Asis_Value : Result(Float_Result);
   begin
      -- Ensure this is an expression
      if Element_Kind(Exp) = An_Expression then
         Asis_Value := E(Exp, Float_Result);
         Value := Asis_Value;
         Valid := Resolved;
      else
         Valid := Improper_Element;
      end if;
   exception
      when Cannot_Evaluate =>
         Valid := No_Value;
      when Asis_Failed =>
         Valid := Asis_Failure;
      when Constraint_Error =>
         Valid := Out_Of_Range;
   end Evaluate_Float_Value_Of;

   --------------------
   -- Float_Value_Of --
   --------------------

   procedure Float_Value_Of (Exp   : in     Asis.Expression;
                             Value :    out Float;
                             Valid :    out Validity) is
      Asis_Value : Result(Float_Result);
   begin
      Evaluate_Float_Value_Of(Exp, Asis_Value, Valid);
      if Valid = Resolved then
         Value := Asis_Value.F_Result;
      end if;
   end Float_Value_Of;

end Asis_Expression_Resolver;
