----------------------------------------------------------------------
--                                                                  --
--                      ASIS_Expression_Resolver                    --
--                                                                  --
--                  Copyright (c) 2001, Marc A. Criley              --
--                                                                  --
-- ASIS_Expression_Resolver is free  software; you can redistribute --
-- it  and/or modify  it  under  terms of  the  GNU General  Public --
-- License  as published  by the  Free Software  Foundation; either --
-- version   2,   or   (at   your  option)   any   later   version. --
-- ASIS_Expression_Resolver is distributed in the hope that it will --
-- be useful,  but WITHOUT ANY  WARRANTY; without even  the implied --
-- warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. --
-- See the GNU General Public License for more details.             --
--                                                                  --
-- As a special exception, if other files instantiate generics from --
-- this unit, or you link this  unit with other files to produce an --
-- executable,  this unit does  not by  itself cause  the resulting --
-- executable  to be  covered by  the GNU  General  Public License. --
-- This exception does not however invalidate any other reasons why --
-- the executable file might be covered by the GNU Public License.  --
--                                                                  --
-- If  you have comments,  questions, have  found bugs  (and better --
-- yet,  bug fixes  :-), or  wish to  suggest  additional features, --
-- please contact the author at mcriley@acm.org.                    --
--                                                                  --
-- Change History:                                                  --
--    1.0           7 Jan 2001   Initial Release                    --
--    2.0           7 Jan 2001   Major additions to support Float   --
--                               and generic discrete types         --
--                                                                  --
-- Known bugs:                                                      --
--    Evaluating a type conversion expression results in an attempt --
-- to resolve the argument of the type conversion as a value of the --
-- requested expression type.  For example, given:                  --
--    N : Natural := Integer(45.2);                                 --
-- Invoking Float_Value_Of on this expression results in a resolved --
-- value of 45.2,  rather than a No_Value (the  type is Natural, so --
-- shouldn't be resolvable as a float).                             --
--                                                                  --
----------------------------------------------------------------------

with Asis;
with System;

package Asis_Expression_Resolver is

   -------------------------------------------------------------------
   -- Validity results:
   --   Resolved         - A numeric value could be determined
   --   No_Value         - No numeric value could be resolved
   --   Improper_Element - The element provided was not of the
   --                         required kind, An_Expression
   --   ASIS_Failure     - An internal ASIS problem occurred
   --   Out_Of_Range     - The expression was resolved, but lay
   --                         outside the allowed value range
   -------------------------------------------------------------------
   type Validity is (Resolved,
                     No_Value,
                     Improper_Element,
                     ASIS_Failure,
                     Out_Of_Range);

   -------------------------------------------------------------------
   -- Calculate the integer value of the given expression if it can
   -- be statically determined.  If the expression resolves to a
   -- character or enumeration value, return a numeric value
   -- equivalent to the 'Pos of that value.
   --
   -- Expressions can be resolved that consist of:
   --    - Discrete literals
   --           95, -1979, False, True, 'A', Ada.Strings.Center
   --    - Constant values
   --           A1       : constant := 10;
   --           Little_Z : constant := 'z';
   --    - Initialized variables (of course only their initial
   --        value can be extracted)
   --           A2 : Natural := 15;
   --           Z1 : Character := Little_Z;
   --    - 'First, 'Last of discrete ranges (see Note below)
   --           Natural'Last, My_Int_Range'First, Character'Last
   --    - Valid numeric expressions
   --           8 * 5, A1 + 16, (A2 - 3), 100 mod 4, abs(-88)
   --    - Qualified and type converted expressions
   --           Natural'(13), Short_Integer(38)
   -------------------------------------------------------------------

   procedure Integer_Value_Of (Exp   : in     Asis.Expression;
                               Value :    out Integer;
                               Valid :    out Validity);

   -------------------------------------------------------------------
   -- Calculate the floating point value of the given expression if it can
   -- be statically determined.
   --
   -- Expressions can be resolved that consist of:
   --    - Discrete literals
   --           95.0, -1979.87
   --    - Constant values
   --           F1 : constant := 10.0;
   --    - Initialized variables (of course only their initial
   --        value can be extracted)
   --           F2 : Float := 15.0;
   --           F3 : My_Float := F1;
   --    - 'First, 'Last of specified ranges (see Note below)
   --           Float'First, My_Float'Last
   --    - Valid numeric expressions
   --           8.0 * 5.0, F1 + 16.6, (F2 - 3.98), abs(-8.8)
   --    - Qualified and some type converted expressions
   --           Float'(13.0), Float(38)
   -------------------------------------------------------------------

   procedure Float_Value_Of (Exp   : in     Asis.Expression;
                             Value :    out Float;
                             Valid :    out Validity);

   -------------------------------------------------------------------
   -- Calculate the cardinal (unsigned) value of the given expression
   -- if it can be statically determined, and interpret it as a member
   -- of the specified discrete scalar type, whether that type be
   -- numeric or an enumeration.
   --
   -- Instantiations of this procedure must be used with care, as
   -- there is no way to verify that the type of the expression to be
   -- resolved corresponds to that of the supplied discrete type.
   --
   -- There are three primary outcomes from invoking this procedure:
   --  I.  The discrete type corresponds to that of the expression,
   --      so all is well.
   --  II. The expression type and discrete types do not
   --      correspond, however the expression value lies within the
   --      cardinality ('Pos values) of the discrete type.  The
   --      literal whose 'Value equals that of the expression result
   --      is returned.  This results in a sort of
   --      quasi-Unchecked_Conversion.
   -- III. The expression type and discrete types do not correspond
   --      and the expression value lies outside the discrete type's
   --      cardinality.  In this case, the Valid parameter is set to
   --      "Out_Of_Range".
   -------------------------------------------------------------------
   generic
      type Cardinal_Type is (<>);
   procedure Cardinal_Value_Of (Exp   : in     Asis.Expression;
                                Value :    out Cardinal_Type;
                                Valid :    out Validity);



   -------------------------------------------------------------------
   -- NOTE: The relationship of 'First and 'Last to subtypes and
   -- derived types with additional constraints is to "unwind" the
   -- subtyping/derivation hierarchy, even if mixed together, until a
   -- constraint is encountered, even if that means going all the way
   -- back to the base type.  Range constraints can be specified using
   -- a simple expression range, e.g., "Min_Val .. Max_Val" or with
   -- the 'Range attribute.
   -------------------------------------------------------------------

end Asis_Expression_Resolver;
